/*

	Astronomical calculation routines

*/

#include "sunclock.h"

/*  Astronomical constants  */

#define epoch	    2444238.5	   /* 1980 January 0.0 */

/*  Constants defining the Sun's apparent orbit  */

#define elonge	    278.833540	   /* Ecliptic longitude of the Sun
				      at epoch 1980.0 */
#define elongp	    282.596403	   /* Ecliptic longitude of the Sun at
				      perigee */
#define eccent      0.016718       /* Eccentricity of Earth's orbit */
#define sunangsiz   0.533128       /* Sun's angular size, degrees, at
				      semi-major axis distance */

/*  Elements of the Moon's orbit, epoch 1980.0  */

#define mmlong      64.975464      /* Moon's mean longitude at the epoch */
#define mmlongp     349.383063	   /* Mean longitude of the perigee at the
				      epoch */
#define mlnode	    151.950429	   /* Mean longitude of the node at the
				      epoch */
#define minc        5.145396       /* Inclination of the Moon's orbit */
#define mecc        0.054900       /* Eccentricity of the Moon's orbit */
#define mangsiz     0.5181         /* Moon's angular size at distance a
				      from Earth */
#define msmax       384401.0       /* Semi-major axis of Moon's orbit in km */
#define mparallax   0.9507	   /* Parallax at distance a from Earth */

#define dsin(x) (sin(dtr((x)))) 			/* Sin from deg */
#define dcos(x) (cos(dtr((x)))) 			/* Cos from deg */

/*  JTIME  --  Convert a Unix date and time (tm) structure to astronomical
			   Julian time (i.e. Julian date plus day fraction,
			   expressed as a double).  */

double DLL jtime(struct tm FAR *t)
{
	double j;
	
    j = ucttoj(t->tm_year + 1900, t->tm_mon, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
	return j;
}

/*  UCTTOJ  --	Convert GMT date and time to astronomical
		Julian time (i.e. Julian date plus day fraction,
		expressed as a double).  */

double DLL ucttoj(long year, int mon, int mday,
	      int hour, int min, int sec)
{

	// Algorithm as given in Meeus, Astronomical Algorithms, Chapter 7, page 61

	int a, b, m;
	long y;
	double j;

    assert(mon  >= 0 && mon  < 12);
    assert(mday >  0 && mday < 32);
    assert(hour >= 0 && hour < 24);
    assert(min  >= 0 && min  < 60);
    assert(sec  >= 0 && sec  < 60);

    m = mon + 1;
    y = year;

	if (m <= 2) {
		y--;
		m += 12;
	}

	/* Determine whether date is in Julian or Gregorian calendar based on
	   canonical date of calendar reform. */

	if ((year < 1582) || ((year == 1582) && ((mon < 9) || (mon == 9 && mday < 5)))) {
		b = 0;
	} else {
		a = ((int) (y / 100));
		b = 2 - a + (a / 4);
	}

	j = (((long) (365.25 * (y + 4716))) + ((int) (30.6001 * (m + 1))) +
				mday + b - 1524.5) +
			((sec + 60L * (min + 60L * hour)) / 86400.0);
	return j;
}

/*  JYEAR  --  Convert	Julian	date  to  year,  month, day, which are
	       returned via integer pointers to integers (note that year is a long).  */

void DLL jyear(double td, long FAR *yy, int FAR *mm, int FAR *dd)
{
	double z, f, a, alpha, b, c, d, e;

	td += 0.5;
	z = floor(td);
	f = td - z;

	if (z < 2299161.0) {
		a = z;
	} else {
		alpha = floor((z - 1867216.25) / 36524.25);
		a = z + 1 + alpha - floor(alpha / 4);
	}

	b = a + 1524;
	c = floor((b - 122.1) / 365.25);
	d = floor(365.25 * c);
	e = floor((b - d) / 30.6001);

	*dd = (int) (b - d - floor(30.6001 * e) + f);
	*mm = (int) ((e < 14) ? (e - 1) : (e - 13));
	*yy = (long) ((*mm > 2) ? (c - 4716) : (c - 4715));
}

/*  JHMS  --  Convert Julian time to hour, minutes, and seconds.  */

void DLL jhms(double j, int FAR *h, int FAR *m, int FAR *s)
{
    long ij;

    j += 0.5;			      /* Astronomical to civil */
    ij = (long) ((j - floor(j)) * 86400.0);
    *h = (int) (ij / 3600L);
    *m = (int) ((ij / 60L) % 60L);
    *s = (int) (ij % 60L);
}

/*  KEPLER  --	Solve the equation of Kepler.  */

static double kepler(double m, double ecc)
{
    double e, delta;
#define EPSILON 1E-6

    e = m = dtr(m);
    do {
	delta = e - ecc * sin(e) - m;
	e -= delta / (1 - ecc * cos(e));
    } while (abs(delta) > EPSILON);
    return e;
}

/*  GMST  --  Calculate Greenwich Mean Siderial Time for a given
	      instant expressed as a Julian date and fraction.	*/

double DLL gmst(double jd)
{
    double t, theta0;

    /* Time, in Julian centuries of 36525 ephemeris days,
       measured from the epoch 1900 January 0.5 ET. */

    t = ((floor(jd + 0.5) - 0.5) - 2415020.0) / JulianCentury;

    theta0 = 6.6460656 + 2400.051262 * t + 0.00002581 * t * t;

    t = (jd + 0.5) - (floor(jd + 0.5));

    theta0 += (t * 24.0) * 1.002737908;

    theta0 = (theta0 - 24.0 * (floor(theta0 / 24.0)));

    return theta0;
}

/*  PHASE  --  Calculate phase of moon as a fraction:

    The  argument  is  the  time  for  which  the  phase is requested,
    expressed as a Julian date and fraction.  Returns  the  terminator
    phase  angle  as a percentage of a full circle (i.e., 0 to 1), and
    stores into pointer arguments  the	illuminated  fraction  of  the
    Moon's  disc, the Moon's age in days and fraction, the distance of
    the Moon from the centre of the Earth, and	the  angular  diameter
    subtended  by the Moon as seen by an observer at the centre of the
    Earth.
*/

double DLL phase(
  double  pdate,		      /* Date for which to calculate phase */
  double  FAR *pphase,		  /* Illuminated fraction */
  double  FAR *mage,		  /* Age of moon in days */
  double  FAR *dist,		  /* Distance in kilometres */
  double  FAR *angdia,		  /* Angular diameter in degrees */
  double  FAR *sudist,		  /* Distance to Sun */
  double  FAR *suangdia)      /* Sun's angular diameter */
{

    double Day, N, M, Ec, Lambdasun, ml, MM, MN, Ev, Ae, A3, MmP,
	   mEc, A4, lP, Varia, lPP, NP, y, x, Lambdamoon,
	   MoonAge, MoonPhase,
	   MoonDist, MoonDFrac, MoonAng,
	   F, SunDist, SunAng;

    /* Calculation of the Sun's position */

    Day = pdate - epoch;		    /* Date within epoch */
    N = fixangle((360 / 365.2422) * Day);   /* Mean anomaly of the Sun */
    M = fixangle(N + elonge - elongp);	    /* Convert from perigee
					       co-ordinates to epoch 1980.0 */
    Ec = kepler(M, eccent);		    /* Solve equation of Kepler */
    Ec = sqrt((1 + eccent) / (1 - eccent)) * tan(Ec / 2);
    Ec = 2 * rtd(atan(Ec));		  /* True anomaly */
    Lambdasun = fixangle(Ec + elongp);      /* Sun's geocentric ecliptic
					       longitude */
    /* Orbital distance factor */
    F = ((1 + eccent * cos(dtr(Ec))) / (1 - eccent * eccent));
    SunDist = SunSMAX / F;		    /* Distance to Sun in km */
    SunAng = F * sunangsiz;                 /* Sun's angular size in degrees */

    /* Calculation of the Moon's position */

    /* Moon's mean longitude */
    ml = fixangle(13.1763966 * Day + mmlong);

    /* Moon's mean anomaly */
    MM = fixangle(ml - 0.1114041 * Day - mmlongp);

    /* Moon's ascending node mean longitude */
    MN = fixangle(mlnode - 0.0529539 * Day);

    /* Evection */
    Ev = 1.2739 * sin(dtr(2 * (ml - Lambdasun) - MM));

    /* Annual equation */
    Ae = 0.1858 * sin(dtr(M));

    /* Correction term */
    A3 = 0.37 * sin(dtr(M));

    /* Corrected anomaly */
    MmP = MM + Ev - Ae - A3;

    /* Correction for the equation of the centre */
    mEc = 6.2886 * sin(dtr(MmP));

    /* Another correction term */
    A4 = 0.214 * sin(dtr(2 * MmP));

    /* Corrected longitude */
    lP = ml + Ev + mEc - Ae + A4;

    /* Variation */
    Varia = 0.6583 * sin(dtr(2 * (lP - Lambdasun)));

    /* True longitude */
    lPP = lP + Varia;

    /* Corrected longitude of the node */
    NP = MN - 0.16 * sin(dtr(M));

    /* Y inclination coordinate */
    y = sin(dtr(lPP - NP)) * cos(dtr(minc));

    /* X inclination coordinate */
    x = cos(dtr(lPP - NP));

    /* Ecliptic longitude */
    Lambdamoon = rtd(atan2(y, x));
    Lambdamoon += NP;

    /* Ecliptic latitude */
//    BetaM = rtd(asin(sin(dtr(lPP - NP)) * sin(dtr(minc))));

    /* Calculation of the phase of the Moon */

    /* Age of the Moon in degrees */
    MoonAge = lPP - Lambdasun;

    /* Phase of the Moon */
    MoonPhase = (1 - cos(dtr(MoonAge))) / 2;

    /* Calculate distance of moon from the centre of the Earth */

    MoonDist = (msmax * (1 - mecc * mecc)) /
	       (1 + mecc * cos(dtr(MmP + mEc)));

    /* Calculate Moon's angular diameter */

    MoonDFrac = MoonDist / msmax;
    MoonAng = mangsiz / MoonDFrac;

    /* Calculate Moon's parallax */

//    MoonPar = mparallax / MoonDFrac;

    *pphase = MoonPhase;
    *mage = SynMonth * (fixangle(MoonAge) / 360.0);
    *dist = MoonDist;
    *angdia = MoonAng;
    *sudist = SunDist;
    *suangdia = SunAng;
    return fixangle(MoonAge) / 360.0;
}

/*  HIGHMOON  --  High-precision moon position calculation.  From Chapter
                  30 of Jean Meeus' Astronomical Formulae for Calculators.
				  Returns celestial longitude and latitude measured along
				  the ecliptic.  */

void DLL highmoon(double jd, double FAR *Lambda, double FAR *Beta)
{
	double t, t2, t3, lprime, m, mprime, d, f, omega, v, e, e2,
	       b, om1, om2;

	/* Time in Julian centuries from 1900 January 0.5 */
	t = (jd - 2415020.0) / JulianCentury;
	t2 = t * t;		   /* Square for frequent use */
	t3 = t2 * t;		   /* Cube for frequent use */

        /* Moon's mean longitude */
	lprime = 270.434164 + 481267.8831 * t - 0.001133 * t2 +
		 0.0000019 * t3;

        /* Sun's mean anomaly */
	m = 358.475833 + 35999.0498 * t - 0.000150 * t2 -
	    0.0000033 * t3;

        /* Moon's mean anomaly */
	mprime = 296.104608 + 477198.8491 * t + 0.009192 * t2 +
		 0.0000144 * t3;

        /* Moon's mean elongation */
	d = 350.737486 + 445267.1142 * t - 0.001436 * t2 +
	    0.0000019 * t3;

	/* Mean distance of the Moon from its ascending node */
	f = 11.250889 + 483202.0251 * t - 0.003211 * t2 -
	    0.0000003 * t3;

        /* Longitude of the Moon's ascending node */
	omega = 259.183275 - 1934.1420 * t + 0.002078 * t2 +
		0.0000022 * t3;

	/* Explicitly range-reduce base angles to reduce additive
	   round-off in subsequent calculations. */

	lprime = fixangle(lprime);
	m = fixangle(m);
	mprime = fixangle(mprime);
	d = fixangle(d);
	f = fixangle(f);
	omega = fixangle(omega);

	/* Additive terms:
		1782-year periodic variations  */
	v = sin(dtr(51.2 + 20.2 * t));
	lprime +=   0.000233 * v;
	m      +=  -0.001778 * v;
	mprime +=   0.000817 * v;
	d      +=   0.002011 * v;

	/*	 The Great Venus Term: 271 year period */
	v = 0.003964 *
	    sin(dtr(346.560 + 132.870 * t - 0.0091731 * t2));
	lprime += v;
	mprime += v;
	d += v;
	f += v;

	/*	 Corrections based on ascending node longitude */
	v = sin(dtr(omega));
	lprime +=  0.001964 * v;
	mprime +=  0.002541 * v;
	d      +=  0.001964 * v;
	f      += -0.024691 * v;
	f      += -0.004328 * sin(dtr(omega + 275.05 - 2.30 * t));

	/* Common multiplicative factors used in the expansions. */

	e = 1 - 0.002495 * t - 0.00000752 * t2;
	e2 = e * e;

	/* Geocentric longitude of the Moon */

	*Lambda = lprime
		    +	   6.288750 * dsin(mprime)
		    +	   1.274018 * dsin(2 * d - mprime)
		    +	   0.658309 * dsin(2 * d)
		    +	   0.213616 * dsin(2 * mprime)
		    - e *  0.185596 * dsin(m)
		    -	   0.114336 * dsin(2 * f)
		    +	   0.058793 * dsin(2 * d - 2 * mprime)
		    + e *  0.057212 * dsin(2 * d - m - mprime)
		    +	   0.053320 * dsin(2 * d + mprime)
		    + e *  0.045874 * dsin(2 * d - m)
		    + e *  0.041024 * dsin(mprime - m)
		    -	   0.034718 * dsin(d)
		    - e *  0.030465 * dsin(m + mprime)
		    +	   0.015326 * dsin(2 * d - 2 * f)
		    -	   0.012528 * dsin(2 * f + mprime)
		    -	   0.010980 * dsin(2 * f - mprime)
		    +	   0.010674 * dsin(4 * d - mprime)
		    +	   0.010034 * dsin(3 * mprime)
		    +	   0.008548 * dsin(4 * d - 2 * mprime)
		    - e *  0.007910 * dsin(m - mprime + 2 * d)
		    - e *  0.006783 * dsin(2 * d + m)
		    +	   0.005162 * dsin(mprime - d)
		    + e *  0.005000 * dsin(m + d)
		    + e *  0.004049 * dsin(mprime - m + 2 * d)
		    +	   0.003996 * dsin(2 * mprime + 2 * d)
		    +	   0.003862 * dsin(4 * d)
		    +	   0.003665 * dsin(2 * d - 3 * mprime)
		    + e *  0.002695 * dsin(2 * mprime - m)
		    +	   0.002602 * dsin(mprime - 2 * f - 2 * d)
		    + e *  0.002396 * dsin(2 * d - m - 2 * mprime)
		    -	   0.002349 * dsin(mprime + d)
		    + e2 * 0.002249 * dsin(2 * d - 2 * m)
		    - e *  0.002125 * dsin(2 * mprime + m)
		    - e2 * 0.002079 * dsin(2 * m)
		    + e2 * 0.002059 * dsin(2 * d - mprime - 2 * m)
		    -	   0.001773 * dsin(mprime + 2 * d - 2 * f)
		    -	   0.001595 * dsin(2 * f + 2 * d)
		    + e *  0.001220 * dsin(4 * d - m - mprime)
		    -	   0.001110 * dsin(2 * mprime + 2 * f)
		    +	   0.000892 * dsin(mprime - 3 * d)
		    - e *  0.000811 * dsin(m + mprime + 2 * d)
		    + e *  0.000761 * dsin(4 * d - m - 2 * mprime)
		    + e2 * 0.000717 * dsin(mprime - 2 * m)
		    + e2 * 0.000704 * dsin(mprime - 2 * m - 2 * d)
		    + e *  0.000693 * dsin(m - 2 * mprime + 2 * d)
		    + e *  0.000598 * dsin(2 * d - m - 2 * f)
		    +	   0.000550 * dsin(mprime + 4 * d)
		    +	   0.000538 * dsin(4 * mprime)
		    + e *  0.000521 * dsin(4 * d - m)
		    +	   0.000486 * dsin(2 * mprime - d);

	/* Geocentric latitude of the Moon */

	b =	     5.128189 * dsin(f)
	      +      0.280606 * dsin(mprime + f)
	      +      0.277693 * dsin(mprime - f)
	      +      0.173238 * dsin(2 * d - f)
	      +      0.055413 * dsin(2 * d + f - mprime)
	      +      0.046272 * dsin(2 * d - f - mprime)
	      +      0.032573 * dsin(2 * d + f)
	      +      0.017198 * dsin(2 * mprime + f)
	      +      0.009267 * dsin(2 * d + mprime - f)
	      +      0.008823 * dsin(2 * mprime - f)
	      + e *  0.008247 * dsin(2 * d - m - f)
	      +      0.004323 * dsin(2 * d - f - 2 * mprime)
	      +      0.004200 * dsin(2 * d + f + mprime)
	      + e *  0.003372 * dsin(f - m - 2 * d)
	      + e *  0.002472 * dsin(2 * d + f - m - mprime)
	      + e *  0.002222 * dsin(2 * d + f - m)
	      + e *  0.002072 * dsin(2 * d - f - m - mprime)
	      + e *  0.001877 * dsin(f - m + mprime)
	      +      0.001828 * dsin(4 * d - f - mprime)
	      - e *  0.001803 * dsin(f + m)
	      -      0.001750 * dsin(3 * f)
	      + e *  0.001570 * dsin(mprime - m - f)
	      -      0.001487 * dsin(f + d)
	      - e *  0.001481 * dsin(f + m + mprime)
	      + e *  0.001417 * dsin(f - m - mprime)
	      + e *  0.001350 * dsin(f - m)
	      +      0.001330 * dsin(f - d)
	      +      0.001106 * dsin(f + 3 * mprime)
	      +      0.001020 * dsin(4 * d - f)
	      +      0.000833 * dsin(f + 4 * d - mprime)
	      +      0.000781 * dsin(mprime - 3 * f)
	      +      0.000670 * dsin(f + 4 * d - 2 * mprime)
	      +      0.000606 * dsin(2 * d - 3 * f)
	      +      0.000597 * dsin(2 * d + 2 * mprime - f)
	      + e *  0.000492 * dsin(2 * d + mprime - m - f)
	      +      0.000450 * dsin(2 * mprime - f - 2 * d)
	      +      0.000439 * dsin(3 * mprime - f)
	      +      0.000423 * dsin(f + 2 * d + 2 * mprime)
	      +      0.000422 * dsin(2 * d - f - 3 * mprime)
	      - e *  0.000367 * dsin(m + f + 2 * d - mprime)
	      - e *  0.000353 * dsin(m + f + 2 * d)
	      +      0.000331 * dsin(f + 4 * d)
	      + e *  0.000317 * dsin(2 * d + f - m + mprime)
	      + e2 * 0.000306 * dsin(2 * d - 2 * m - f)
	      -      0.000283 * dsin(mprime + 3 * f);

	om1 = 0.0004664 * dcos(omega);
	om2 = 0.0000754 * dcos(omega + 275.05 - 2.30 * t);
	*Beta = b * (1 - om1 - om2);
}

/*  OBLIQEQ  --  Calculate the obliquity of the ecliptic for a given Julian
				 date.  This uses Laskar's tenth-degree polynomial fit
				 (J. Laskar, Astronomy and Astrophysics, Vol. 157, page 68 [1986])
				 which is accurate to within 0.01 arc second between AD 1000
				 and AD 3000, and within a few seconds of arc for +/-10000
				 years around AD 2000.  If we're outside the range in which
				 this fit is valid (deep time) we simply return the J2000 value
				 of the obliquity, which happens to be almost precisely the mean.  */

double DLL obliqeq(double jd)
{
#define Asec(x)	((x) / 3600.0)

	static double oterms[10] = {
		Asec(-4680.93),
		Asec(   -1.55),
		Asec( 1999.25),
		Asec(  -51.38),
		Asec( -249.67),
		Asec(  -39.05),
		Asec(    7.12),
		Asec(   27.87),
		Asec(    5.79),
		Asec(    2.45)
	};

	double eps = 23 + (26 / 60.0) + (21.448 / 3600.0), u, v;
	int i;

	v = u = (jd - J2000) / (JulianCentury * 100);

    if (abs(u) < 1.0) {
		for (i = 0; i < 10; i++) {
			eps += oterms[i] * v;
			v *= u;
		}
	}
	return eps;
}

/*  ECLIPTOEQ  --  Convert celestial (ecliptical) longitude and latitude into
				   right ascension (in degrees) and declination.  We must supply
				   the time of the conversion in order to compensate correctly
				   for the varying obliquity of the ecliptic over time.  */

void DLL ecliptoeq(double jd, double Lambda, double Beta,
					   double FAR *Ra, double FAR *Dec)
{
	double eps;

    /* Obliquity of the ecliptic. */

    eps = dtr(obliqeq(jd));

    *Ra = fixangle(rtd(atan2((cos(eps) * sin(dtr(Lambda)) -
					     (tan(dtr(Beta)) * sin(eps))), cos(dtr(Lambda)))));
    *Dec = rtd(asin((sin(eps) * sin(dtr(Lambda)) * cos(dtr(Beta))) +
			     (sin(dtr(Beta)) * cos(eps))));

}

/*  DEFINEPRECESSION, PRECESSOBJECT  --  Precess equatorial coordinates from J2000.0
										 to a given target epoch specified as a year and
										 optional fraction. */

//	D'aprs Meeus, Astronomical Algorithms, Chapter 20, Rigorous method: page 126

static double preZeta, preZ, preTheta;

void DLL definePrecession(double targetEpoch)
{
	double t, t2, t3;

	t = (targetEpoch - 2000.0) / 100.0;
	t3 = (t2 = t * t) * t;
#define SecToR(x)	(dtr((x)) / 3600.0)	// Seconds of arc to radians
	preZeta = SecToR(2306.2181 * t + 0.30188 * t2 + 0.017998 * t3);
	preZ = SecToR(2306.2181 * t + 1.09468 * t2 + 0.018203 * t3);
	preTheta = SecToR(2004.3109 * t - 0.42665 * t2 - 0.041833 * t3);
}

void DLL precessObject(double ira, double idec, double FAR *ora, double FAR *odec)
{
	double rira = dtr(ira), ridec = dtr(idec), a, b, c;

	a = cos(ridec) * sin(rira + preZeta);
	b = cos(preTheta) * cos(ridec) * cos(rira + preZeta) - sin(preTheta) * sin(ridec);
	c = sin(preTheta) * cos(ridec) * cos(rira + preZeta) + cos(preTheta) * sin(ridec);

	*ora = rtd(atan2(a, b) + preZ);
	*odec = rtd((idec > 85.0) ? acos(sqrt(a * a + b * b)) : asin(c));
}
